from sklearn.neighbors import KDTree

class NearestNeighbors:
    """Class to create KDTree and query the Nearest Neighbor (NN) nodes

    """
    def __init__(self, nodeCoordArray, leafSize):
        self.tree = KDTree(nodeCoordArray, leaf_size=leafSize)

    def getSKnode(self, tree, nodeCoordArray, kx, nodeNumber):
        """K-based NN query for the specific collocation point. #<kr> K-based 특정 노드에 대한 NN 쿼리.

        Parameters
            - tree - :class:`sklearn.neighbors._kd_tree.KDTree`. Sci-kit learn-based KDTree for fast generalized N-point problems.
            - nodeCoordArray - :class:`numpy.ndarray`. node coordinate. #<kr> nodeCoordArray는 좌표값이 순서대로 들어간 것이다. gmsh에서는 node가 1부터 순서대로 생성되기 때문에. nodeCoordArray의 index에 1을 더하면 gmsh의 node tag와 같은 값이 된다.
            - kx - :class:`int`. The number of nearest neighbors to return.
            - nodeNumber - :class:`int`. Tag number.
        
        Returns
            - R1 (distance) - :class:`numpy.ndarray`. #<kr> 하나의 점을 기준으로, K기반 연결점(들)(nearest neighbors)간의 거리.
            - R2 (nodeTagMinusOne) - :class:`numpy.ndarray`. #<kr> 하나의 점을 기준으로, 이들과 연결된 점(들)(nearest neighbors) id. 이 점들의 id는 gmsh에서 나타나는 node tag1 보다 1이 작다.
        """
        distance, nodeTagMinusOne = tree.query(nodeCoordArray[nodeNumber-1:nodeNumber], k=kx, dualtree=True)
        return distance, nodeTagMinusOne


    def getSRNode(self, tree, nodeCoordArray, radius, nodeNumber):
        #region doc. string
        """R-based NN query for the specific collocation point. #<kr> Radius-based 특정 노드에 대한 NN 쿼리.

        Parameters
            - tree - :class:`sklearn.neighbors._kd_tree.KDTree`. Sci-kit learn-based KDTree for fast generalized N-point problems.
            - nodeCoordArray - :class:`numpy.ndarray`. node coordinate. #<kr> nodeCoordArray는 좌표값이 순서대로 들어간 것이다. gmsh에서는 node가 1부터 순서대로 생성되기 때문에. nodeCoordArray의 index에 1을 더하면 gmsh의 node tag와 같은 값이 된다.
            - radius - :class:`float`. Distance within which neighbors are returned.
            - nodeNumber - :class:`int`. Tag number.
        
        Returns
            - R1 (distance) - :class:`numpy.ndarray`. #<kr> 하나의 점을 기준으로, R기반 연결점(들)(nearest neighbors)간의 거리.
            - R2 (nodeTagMinusOne) - :class:`numpy.ndarray`. #<kr> 하나의 점을 기준으로, 이들과 연결된 점(들)(nearest neighbors) id. 이 점들의 id는 gmsh에서 나타나는 node tag1 보다 1이 작다.
        """
        #endregion doc. string
        nodeTagMinusOne, distance = tree.query_radius(nodeCoordArray[nodeNumber-1:nodeNumber], r=radius, return_distance=True, sort_results=True)
        return distance, nodeTagMinusOne

    
    #<kr> Python은 일반적인 다른 랭귀지에서의 method overload가 안됨. 변수를 가변적으로 받을 수 있어서 그런듯.
    def getNodes(self, tree, nodeCoordArray, var):
        #region doc. string
        """Get the nearest neighbor (NN) nodes.

        - If var is an integer type, K-based NNs are returned.
        - If var is a float type, radius-based NNs are returned.

        Parameters
            - tree - :class:`sklearn.neighbors._kd_tree.KDTree`. Sci-kit learn-based KDTree for fast generalized N-point problems.
            - nodeCoordArray - :class:`numpy.ndarray`. node coordinate. #<kr> nodeCoordArray는 좌표값이 순서대로 들어간 것이다. gmsh에서는 node가 1부터 순서대로 생성되기 때문에. nodeCoordArray의 index에 1을 더하면 gmsh의 node tag와 같은 값이 된다.
            - var - :class:`int` or :class:`float`. Query parameter of the KD tree. 
        
        Returns
            - R1 (distance) - :class:`numpy.ndarray`.
            - R2 (nodeTagMinusOne) - :class:`numpy.ndarray`. #<kr> 하나의 점을 기준으로, 이들과 연결된 점(들)(nearest neighbors) id. 이 점들의 id는 gmsh에서 나타나는 node tag1 보다 1이 작다.
        """
        #endregion doc. string

        if type(var) is int:
            distance, nodeTagMinusOne = self.__getNodes_int(tree, nodeCoordArray, var) # var = k
            return distance, nodeTagMinusOne
        elif type(var) is float:
            distance, nodeTagMinusOne = self.__getNodes_flot(tree, nodeCoordArray, var) # var = radius
            return distance, nodeTagMinusOne

        else:
            print("neither")

    # private function
    def __getNodes_int(self, tree, nodeCoordArray, var): # var = k
        #region doc. string
        """Get the K-based nearest neighbor (NN) nodes.

        Parameters
            - tree - :class:`sklearn.neighbors._kd_tree.KDTree`. Sci-kit learn-based KDTree for fast generalized N-point problems.
            - nodeCoordArray - :class:`numpy.ndarray`. node coordinate. nodeCoordArray는 좌표값이 순서대로 들어간 것이다. gmsh에서는 node가 1부터 순서대로 생성되기 때문에. nodeCoordArray의 index에 1을 더하면 gmsh의 node tag와 같은 값이 된다.
            - var - :class:`int`. The number of nearest neighbors to return. 
        
        Returns
            - R1 (distance) - :class:`numpy.ndarray`.
            - R2 (nodeTagMinusOne) - :class:`numpy.ndarray`. #<kr> 하나의 점을 기준으로, 이들과 연결된 점(들)(nearest neighbors) id. 이 점들의 id는 gmsh에서 나타나는 node tag1 보다 1이 작다.
        """
        #endregion doc. string

        #<kr> nodeCoordArray는 좌표값이 순서대로 들어간 것이다. gmsh에서는 node가 1부터 순서대로 생성되기 때문에
        #<kr> 일단은 nodeCoordArray의 index에 1을 더하면 gmsh의 node tag와 같은 값이 된다.
        #<kr> nodeTagMinusOne은 gmsh의 node tag와 비교해서 1이 작다는 것을 알리기 위해서 이름을 이렇게 지은것.
        distance, nodeTagMinusOne = tree.query(nodeCoordArray[:], k=var, dualtree=True)
        
        return distance, nodeTagMinusOne

        
    def __getNodes_flot(self, tree, nodeCoordArray, var): # var = radius
        #region doc. string
        """Get the radius-based nearest neighbor (NN) nodes.

        Parameters
            - tree - :class:`sklearn.neighbors._kd_tree.KDTree`. Sci-kit learn-based KDTree for fast generalized N-point problems.
            - nodeCoordArray - :class:`numpy.ndarray`. node coordinate. #<kr> nodeCoordArray는 좌표값이 순서대로 들어간 것이다. gmsh에서는 node가 1부터 순서대로 생성되기 때문에. nodeCoordArray의 index에 1을 더하면 gmsh의 node tag와 같은 값이 된다.
            - var - :class:`float`. Distance within which neighbors are returned. 
        
        Returns
            - R1 (distance) - :class:`numpy.ndarray`.
            - R2 (nodeTagMinusOne) - :class:`numpy.ndarray`. #<kr> 하나의 점을 기준으로, 이들과 연결된 점(들)(nearest neighbors) id. 이 점들의 id는 gmsh에서 나타나는 node tag1 보다 1이 작다.
        """
        #endregion doc. string
        nodeTagMinusOne, distance = tree.query_radius(nodeCoordArray[:], r=var, return_distance=True, sort_results=True)

        return distance, nodeTagMinusOne

        
        
        
        
        



import sys
import gmsh
import numpy
from numpy.core.defchararray import not_equal

class GeometryController:
    #region doc. string
    """Class to control geometric information of the Gmsh model.
    """
    #endregion doc. string
    def __init__(self, fileName=None, dim=3):
        
        
        if not fileName is None:
            # initialize gmsh
            gmsh.initialize(sys.argv)
            gmsh.open(fileName)
            print("File name: " + fileName)
            print("Model name: " + gmsh.model.getCurrent())
            # generate n-dimensional meshes
            gmsh.model.mesh.generate(dim)
        
        #region entire object - attributes
        self.__nodeTags = None
        self.__nodesCoordinates = None
        self.__nodesOnBoundaries = None
        self.__meshTags = None
        self.__boundaryTags = None
        self.__physicalGroups = None
        

        self.__connection_nodeAndCoords = None
        self.__connection_boundaryAndNodes = None
        self.__connection_meshAndNodes = None
        #endregion entire object - attributes

        #region object by physical group - attributes
        self.__connection_physicalGroupAndNodes = None
        self.__connection_physicalGroupAndNormals = None
        #endregion object by physical group - attributes
        self.__connection_nodesAndIntegratedNormals = None

        self.__isGeometricInfo = False

        self.__genGeometricInfo()
        if not fileName is None: gmsh.finalize()

    def __genGeometricInfo(self):
        #region doc. string
        """Function to get all of the geometry-related information.
        
        Returns
            - void
        """
        #endregion doc. string

        #region entire object - method
        nodeTags, nodeCoords, _ = gmsh.model.mesh.getNodes()
        
        nodeAndCoord = {}
        for i in range(0, len(nodeTags)):
            coordTuple = tuple(nodeCoords[i*3:i*3+3])
            if not nodeTags[i] in nodeAndCoord:
                nodeAndCoord[nodeTags[i]] = coordTuple
            else:
                nodeAndCoord[nodeTags[i]].append(coordTuple)

        # get connection between boundary and nodes
        boundaryTags, boundaryMesh_nodes = gmsh.model.mesh.getElementsByType(1)
        connection_boundaryAndNodes, nodesOnBoundaries = self.__getConnection(2, boundaryMesh_nodes, boundaryTags)

        # get connection between mesh and node
        meshTags, triMesh_nodes = gmsh.model.mesh.getElementsByType(2)
        connection_meshAndNodes, nodesOnMeshes = self.__getConnection(3, triMesh_nodes, meshTags)
        
        self.__nodeTags = nodeTags
        #self._nodeCoordinates = Coords(nodeCoords.reshape((-1,3)))
        self.__nodesCoordinates = nodeCoords.reshape((-1,3))
        self.__connection_nodeAndCoords = nodeAndCoord

        self.__boundaryTags = boundaryTags
        self.__connection_boundaryAndNodes = connection_boundaryAndNodes
        self.__nodesOnBoundaries = nodesOnBoundaries

        self.__meshTags = meshTags
        self.__connection_meshAndNodes = connection_meshAndNodes

        #endregion entire object - method

        #region object by physical group - method
        physicalGroups = []
        connection_physicalGroupAndNodes = {}
        connection_physicalGroupAndNormals = {}
        gt2 = gmsh.model.getPhysicalGroups(2)
        # print(gt2)
        # print(gmsh.model.getEntities(2))
        

        for tag in gt2:
            # surfX = tag[1] #tag[0] = dimension, tag[1] = object tag
            # physicalGroupName = gmsh.model.getPhysicalName(2, surfX)
            # physicalGroups.append(physicalGroupName)
            # nodes4PhysicalGroup, _ = gmsh.model.mesh.getNodesForPhysicalGroup(2, surfX)
            # connection_physicalGroupAndNodes[physicalGroupName] = nodes4PhysicalGroup
            physicalGroupName = gmsh.model.getPhysicalName(tag[0], tag[1])
            physicalGroups.append(physicalGroupName)
            nodes4PhysicalGroup, _ = gmsh.model.mesh.getNodesForPhysicalGroup(tag[0], tag[1])
            connection_physicalGroupAndNodes[physicalGroupName] = nodes4PhysicalGroup


        
        # IMPORTANT: We have to use 'parametric coord' to get the normal vector 
        # of the collocation points according to the surface. 
        # We can get it using the 'gmsh.model.mesh.getNodes' method. 
        # But sometimes, this method returns the wrong physical group name. 
        # In the 2D surface model case, it seems that name returns correctly, 
        # but the 3D volume case returns the wrong name. 
        # I think I don't understand this method of the Gmsh API properly, or it might be a bug. 
        # So I reassigned the physical group name using the below process. 
        # However, this is a temporary step, so that it may contain some errors.
        #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # ➡︎ *It seems that I fix it.*
        # for tag in gt2:
            # surf = tag[1] #tag[0] = dimension, tag[1] = object tag
            #<kr> physical group에서 가져오는 tag와 entity tag는 다름. 즉, ⬆︎의 surf는 physical group에서 가져온 tag고, ⬇︎ surf는 entity tag임.
            #<kr> 그림을 포함한 좀 더 자세한 설명은 obsidian://open?vault=Primary&file=CodeTips%2Fgmsh%20entity%20tag
            surf = (gmsh.model.getEntitiesForPhysicalGroup(tag[0], tag[1]))[0]
            nTags, _, paramCoord = gmsh.model.mesh.getNodes(2, surf, True)
            # realPhysicalGroupName = ""
            # for k, v in connection_physicalGroupAndNodes.items():
            #     if sorted(nTags) == sorted(v):
            #         realPhysicalGroupName = k
            #         break

            connection_nodeAndNormals = {}
            
            normals = gmsh.model.getNormal(surf, paramCoord)
            tagIndex = 0
            #for i in nodes4PhysicalGroup:
            for i in nTags:
                temp = [normals[tagIndex], normals[tagIndex+1], normals[tagIndex+2]]
                connection_nodeAndNormals[i] = temp
                tagIndex += 3
                
            connection_physicalGroupAndNormals[physicalGroupName] = connection_nodeAndNormals
            #connection_physicalGroupAndNormals[realPhysicalGroupName] = connection_nodeAndNormals
            
        gt3 = gmsh.model.getPhysicalGroups(3)
        for tag in gt3:
            # physicalGroupName = gmsh.model.getPhysicalName(3, tag[1])
            physicalGroupName = gmsh.model.getPhysicalName(tag[0], tag[1])
            #⬇︎ There are not normal vector of the 3D physical group
            #physicalGroups.append(physicalGroupName)
            # nodes4PhysicalGroup, coord4PhysicalGroup = gmsh.model.mesh.getNodesForPhysicalGroup(3, tag[1])
            nodes4PhysicalGroup, coord4PhysicalGroup = gmsh.model.mesh.getNodesForPhysicalGroup(tag[0], tag[1])
            connection_physicalGroupAndNodes[physicalGroupName] = nodes4PhysicalGroup
        
        self.__physicalGroups = physicalGroups
        self.__connection_physicalGroupAndNodes = connection_physicalGroupAndNodes
        self.__connection_physicalGroupAndNormals = connection_physicalGroupAndNormals
        
        #endregion object by physical group - method

        #region - integrated normal vector
        nodesAndIntegratedNormals = {}
        for i in self.__physicalGroups:
            dicPandB = self.__connection_physicalGroupAndNormals[i]
            for p, b in dicPandB.items():
                if p in nodesAndIntegratedNormals:
                    temp = numpy.array(b) + numpy.array(nodesAndIntegratedNormals[p])
                    nodesAndIntegratedNormals[p] = (temp / numpy.linalg.norm(temp)).tolist()
                else:
                    nodesAndIntegratedNormals[p] = b
        self.__connection_nodesAndIntegratedNormals = nodesAndIntegratedNormals
            
        #endregion - integrated normal vector


        self.__isGeometricInfo = True
        

    def getNodeTags(self, sort=False):
        #region doc. string
        """Function to get all node tags of meshes.

        Parameters
            - sort - :class:`bool`. Sorting tags.
        
        Returns
            - node tags - :class:`numpy.ndarray<int>`. If sort == False.
            - node tags - :class:`list<int>`. If sort == True.
        """
        #endregion doc. string
        if (sort):
            return sorted(self.__nodeTags)
        
        return self.__nodeTags
        
    def getNodesCoordinates(self):
        #region doc. string
        """Function to get all coordinates of the nodes.
        
        Returns
            - node coordinates - :class:`numpy.ndarray<numpy.ndarray<double>>`. sorted.
        """
        #endregion doc. string
        return self.__nodesCoordinates

    def getConnectionBetweenNodeAndCoords(self, nodeTag=None):
        #region doc. string
        """Function to generate the connection between nodes and coordinates.

        Parameters
            - nodeTag - :class:`int`. Tag of a node.
        
        Returns
            - connection - :class:`dict{numpy.uint64, tuple<double>}`. sorted. If nodeTag == None. 
            - coordinates - :class:`tuple<double>`. If nodeTag != None
        """
        #endregion doc. string
        if (nodeTag==None):
            return self.__connection_nodeAndCoords
        else:
            return self.__connection_nodeAndCoords[nodeTag]
    

    def getBoundaryTags(self, sort=False):
        #region doc. string
        """Function to get 1D element tags on the boundaries.

        Parameters
            - sort - :class:`bool`. sort option.
        
        Returns
            - boundary tags - :class:`numpy.ndarray<int>`. If sort == False.
            - boundary tags - :class:`list<int>`. If sort == True.
        """
        #endregion doc. string
        if (sort):
            return sorted(self.__boundaryTags)
        return self.__boundaryTags

    def getNodesOnBoundaries(self):
        #region doc. string
        """Function to get all collocation point tags of boundaries.
        
        Returns
            - 1D element tags on boundaries - :class:`list<int>`.
        """
        #endregion doc. string
        return self.__nodesOnBoundaries


    def getMeshTags(self, sort=False):
        #region doc. string
        """Function to get all mesh tags.

        Parameters
            - sort - :class:`bool`. sort option.
        
        Returns
            - mesh tags - :class:`numpy.ndarray<int>`. If sort == True.
            - mesh tags - :class:`list<int>`. If sort == False.
        """
        #endregion doc. string
        if (sort):
            return sorted(self.__meshTags)
        return self.__meshTags
    
    def getConnectionBetweenMeshAndNodes(self, meshTag=None):
        #region doc. string
        """Function to generate a connection between mesh and nodes.

        Parameters
            - meshTag - :class:`int`. Tag of a mesh.
        
        Returns
            - connection - :class:`dict{Int, tuple<int>}`. If nodeTag == None.
            - node tags - :class:`tuple<int>`. If nodeTag != None.
        """
        #endregion doc. string
        if (meshTag==None):
            return self.__connection_meshAndNodes
        return self.__connection_meshAndNodes[meshTag]

    
    def getConnectionBetweenPhysicalGroupAndNodes(self, physicalGroupName=None, sort=False):
        #region doc. string
        """Function to get a connection between physical group (2-, 3-dimension) and node tags.

        - If you want to get all physical group names, use `dict.keys().`

        Parameters
            - physicalGroupName - :class:`String`.
            - sort - :class:`bool`.
        
        Returns
            - connections - :class:`dict{String, numpy.ndarray<int>}`. If physicalGroupName == None.
            - nodes - :class:`numpy.ndarray<int>`.  If physicalGroupName != None, and sort == False.
            - nodes - :class:`list<int>`.  If physicalGroupName != None, and sort == True.
        """
        #endregion doc. string
        if (physicalGroupName==None):
            return self.__connection_physicalGroupAndNodes
        else:
            if (physicalGroupName in self.__connection_physicalGroupAndNodes):
                if (sort):
                    return sorted(self.__connection_physicalGroupAndNodes[physicalGroupName])
                return self.__connection_physicalGroupAndNodes[physicalGroupName]
            print("************* error *************\n" + "There's no physical name '" + physicalGroupName + "' in the model.\n" + "*********************************\n")


    def getConnectionBetweenPhysicalGroupAndNormals(self, physicalGroupName=None, integration=False, nodeTag=-1):
        #region doc. string
        """Function to get a connection between physical group (2-dimension) and the normal vector.

        Parameters
            - physicalGroupName - :class:`String`.
            - nodeTag - :class:`int`.
        
        Returns
            - connections - :class:`dict{String, list<double>}`. If physicalGroupName == None.
            - normal vector - :class:`list<int>`.  If physicalGroupName != None, and node tag is valid.
        """
        #endregion doc. string
        if (physicalGroupName==None):
            if (not integration):
                    return self.__connection_physicalGroupAndNormals
            else:
                if (nodeTag in self.__nodeTags):
                    return self.__connection_nodesAndIntegratedNormals[nodeTag]
                return self.__connection_nodesAndIntegratedNormals
        else:
            if (physicalGroupName in self.__connection_physicalGroupAndNodes):
                if (nodeTag in self.__connection_physicalGroupAndNodes[physicalGroupName]):
                    return self.__connection_physicalGroupAndNormals[physicalGroupName][nodeTag]
                print("************* error *************\n" + "There's no node No. '" + str(nodeTag) + "' in '" + physicalGroupName + "' in the model.\n" + "*********************************\n")
                return self.__connection_physicalGroupAndNodes[physicalGroupName]
            print("************* error *************\n" + "There's no physical name '" + physicalGroupName + "' in the model.\n" + "*********************************\n")
    

    def getPhysicalGroups(self):
        #region doc. string
        """Function to get the names of the 2D physical groups.

        Parameters
            - void.
        
        Returns
            - physicalGroupName - :class:`str`.
        """
        #endregion doc. string
        return self.__physicalGroups
    

    def __getConnection(self, connectedNodeDim:int, nodes, meshes):
        #region doc. string
        """Connections between boundary or triangular mesh and nodes.

        - EX) Connection between Boundary edge tag and Node tags - {82: (1, 13), 83: (13, 14),...}
        - EX) Connection between Triangular element tag and Node tags - {7: (6, 21, 38), 8: (34, 40, 51), 9: (17, 40, 51),...}

        Parameters
            - connectedNodeDim - :class:`int`. The number of vertices of the object to generate connections. boundary = 2, triangle = 3.
            - nodes - :class:`list<int>`. [Description].
            - meshes - :class:`list<int>`. The tag list of objects for generating connections.
        
        Returns
            - connections - :class:`Dict{Int, List<Int>}`. [Description]
            - nodes on objects - :class:`List`. [Description]
        """
        #endregion doc. string
        connections = {}
        objects = []
        nodesOnObjects = []
        for i in range(0, len(nodes), connectedNodeDim):
            nodesOnObjects.extend(nodes[i:i+connectedNodeDim])
            nodesOnObjects = list(set(nodesOnObjects))
            nodeTuple = tuple(sorted(nodes[i:i+connectedNodeDim]))
            objects.append(nodeTuple)
            elementTag = meshes[int(i/connectedNodeDim)]
            
            if not elementTag in connections:
                connections[elementTag] = nodeTuple
            else:
                connections[elementTag].append(nodeTuple)

        return connections, nodesOnObjects



    def gmshClose(self):
        gmsh.finalize()




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gmshgeometry", # 폴더명과 동일한 이름
    version="0.0.7",
    author="CM2Lab",
    author_email="s.park@openest.eu",
    description="gmsh-based geometry controller",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/cm2lab/geodata_module",
    # ⬇︎ 배포하는 패키지의 다운로드 url을 적어줍니다.
    # ex_ download_url='https://github.com/doorBW/pypi_deploy_test/archive/master.zip',
    download_url='',
    # ⬇︎ 해당 패키지를 사용하기 위해 필요한 패키지를 적어줍니다. ex. install_requires= ['numpy', 'django']
    # 여기에 적어준 패키지는 현재 패키지를 install할때 함께 install됩니다.
    install_requires=[],
    packages=setuptools.find_packages(),
    # 패키지의 키워드를 적습니다.
    keywords=['gmsh geometry'],
    # ⬇︎ 파이썬 파일이 아닌 다른 파일을 포함시키고 싶다면 package_data에 포함시켜야 합니다.
    # ex) package_data= {'pyquibase' : ['db-connectors/sqlite-jdbc-3.18.0.jar', 'db-connectors/mysql-connector-java-5.1.42-bin.jar', 'liquibase/liquibase.jar' ]},
    package_data= {},
    # ⬇︎ 위의 package_data에 대한 설정을 하였다면 zip_safe설정도 해주어야 합니다. package_data가 있다면 zip_safe는 반드시 False로.
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)